import math
import numbers

class Polynome:
    def __init__(self, coefficients):
        coefficients = list(coefficients)
        for i in range(len(coefficients)):
            if coefficients[-1] == 0:
                coefficients.pop(-1)
            else:
                break
        if not coefficients:
            coefficients = (0,) # si aucun coefficient est non vide, on ajoute le coefficient nul à la liste.
        self._coefficients = tuple(coefficients) # mieux vaut stocker la liste des coefficients de manière non mutable.
        # Ça permet d'éviter des erreurs, un polynôme n'a pas de raison de changer de coefficients (c'est un objet non mutable).
        # On met un "_" devant le champ pour laisser définir une propriété, c'est une conventions
    
    @property
    def degree(self):
        if self.coefficients == (0,):
            return -math.inf # le polynôme nul est de degré -∞ 
        return len(self.coefficients) - 1

    @property
    def coefficients(self):
        return self._coefficients

    def __repr__(self):
        s = f"{self.coefficients[0]}"
        if self.degree >= 1:
            c = self.coefficients[1]
            sign = " + "
            if c<0:
                sign = " - "
                c = -c
            s += f"{sign}{c}X"
        i = 2
        for c in self.coefficients[2:]:
            sign = " + "
            if c<0:
                sign = " - "
                c = -c
            s += f"{sign}{c}X^{i}"
            i += 1
        return s

    def __add__(self, other):
        if isinstance(other, numbers.Number):
            other = Polynome([other]) # other est un nombre, on le transforme en Polynôme.
        if len(other.coefficients) > len(self.coefficients):
            return other + self
        coefficients = []
        for i in range(len(self.coefficients)):
            c = self.coefficients[i]
            if i < len(other.coefficients):
                c += other.coefficients[i]
            coefficients.append(c)
        return Polynome(coefficients)

    def __neg__(self):
        return Polynome([-a for a in self.coefficients])

    def __sub__(self, other):
        return self + (-other) # On utilise add et neg pour définir sub

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            other = Polynome([other])
        Q = Polynome([0]) # On va accumuler le résultat dans Q
        for i, c in enumerate(other.coefficients):
            coefficients = [0]*i + [c*e for e in self.coefficients] # On construit les P*cX^i avec P=self.
            Q = Q + Polynome(coefficients) # Et on additionne
        return Q

    def __call__(self, x):
        return sum(map(lambda c,i: c*x**i, self.coefficients, range(self.degree)))

    def __rmul__(self, other):
        return self * other # Ici on a appeler other * self, ce qui est traduit pas self.__rmul__(other) si rmul est défini.

    def __rsub__(self, other):
        return (-self) + other

    def __radd__(self, other):
        return self + other
    
    def derivate(self):
        coefficients = []
        for i, c in enumerate(self.coefficients[1:]):
            coefficients.append(c*(i+1))
        return Polynome(coefficients)
